import os.path
import pytest
import tempfile
import shutil


@pytest.yield_fixture
def tempdir():
    directory = tempfile.mkdtemp()
    yield directory
    shutil.rmtree(directory)


@pytest.fixture
def current_path():
    return os.path.dirname(os.path.realpath(__file__))


@pytest.fixture
def fixtures_path(current_path):
    return os.path.join(current_path, 'fixtures')



import morepath
import os.path

from more.webassets import WebassetsApp
from more.webassets.directives import Asset


def test_webasset_path(current_path):

    current_path = os.path.dirname(os.path.realpath(__file__))

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return './fixtures'

    @App.webasset_path()
    def get_overlapping_path():
        return os.path.join(current_path, './fixtures')

    @App.webasset_path()
    def get_current_path():
        return '.'

    @App.webasset_path()
    def get_parent_path():
        return '..'

    morepath.commit(App)

    app = App()

    assert app.config.webasset_registry.paths == [
        os.path.normpath(os.path.join(current_path, '..')),
        os.path.normpath(os.path.join(current_path, '.')),
        os.path.normpath(os.path.join(current_path, 'fixtures')),
        os.path.normpath(os.path.join(current_path, 'fixtures')),
    ]


def test_webasset_relative(current_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_relative_path():
        return 'fixtures'

    morepath.commit(App)

    assert App().config.webasset_registry.paths == [
        os.path.join(current_path, 'fixtures')
    ]


def test_webasset_path_inheritance(tempdir, current_path):

    os.mkdir(os.path.join(tempdir, 'A'))
    os.mkdir(os.path.join(tempdir, 'B'))
    os.mkdir(os.path.join(tempdir, 'C'))

    class A(WebassetsApp):
        pass

    @A.webasset_path()
    def get_path_a():
        return os.path.join(tempdir, 'A')

    class B(WebassetsApp):
        pass

    @B.webasset_path()
    def get_path_b():
        return os.path.join(tempdir, 'B')

    class C(B, A):
        pass

    @C.webasset_path()
    def get_path_c():
        return os.path.join(tempdir, 'C')

    class D(A, B):
        pass

    @D.webasset_path()
    def get_path_c_2():
        return os.path.join(tempdir, 'C')

    # the order of A and B is defined by the order they are scanned with
    morepath.commit(A, B, C, D)

    assert C().config.webasset_registry.paths == [
        os.path.join(tempdir, 'C'),
        os.path.join(tempdir, 'B'),
        os.path.join(tempdir, 'A'),
    ]

    assert D().config.webasset_registry.paths == [
        os.path.join(tempdir, 'C'),
        os.path.join(tempdir, 'B'),
        os.path.join(tempdir, 'A'),
    ]


def test_webasset_filter():

    class Base(WebassetsApp):
        pass

    @Base.webasset_filter('js')
    def get_base_js_filter():
        return 'jsmin'

    class App(WebassetsApp):
        pass

    @App.webasset_filter('js')
    def get_js_filter():
        return 'rjsmin'

    morepath.commit(App)

    assert App().config.webasset_registry.filters == {'js': 'rjsmin'}


def test_webasset_directive(tempdir, fixtures_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return fixtures_path

    @App.webasset_output()
    def get_output_path():
        return tempdir

    @App.webasset('common')
    def get_common_assets():
        yield 'jquery.js'
        yield 'underscore.js'

    morepath.commit(App)

    app = App()

    assert app.config.webasset_registry.assets == {
        'jquery.js': Asset(
            name='jquery.js',
            assets=(os.path.join(fixtures_path, 'jquery.js'), ),
            filters={}
        ),
        'underscore.js': Asset(
            name='underscore.js',
            assets=(os.path.join(fixtures_path, 'underscore.js'), ),
            filters={}
        ),
        'common': Asset(
            name='common',
            assets=('jquery.js', 'underscore.js'),
            filters={}
        )
    }

    common = list(app.config.webasset_registry.get_bundles('common'))

    assert len(common) == 1
    assert common[0].output.endswith('common.bundle.js')
    assert common[0].contents == (
        os.path.join(fixtures_path, 'jquery.js'),
        os.path.join(fixtures_path, 'underscore.js')
    )

    jquery = list(app.config.webasset_registry.get_bundles('jquery.js'))

    assert len(jquery) == 1
    assert jquery[0].output.endswith('jquery.js.bundle.js')
    assert jquery[0].contents == (
        os.path.join(fixtures_path, 'jquery.js'),
    )

    underscore = list(
        app.config.webasset_registry.get_bundles('underscore.js'))

    assert len(underscore) == 1
    assert underscore[0].output.endswith('underscore.js.bundle.js')
    assert underscore[0].contents == (
        os.path.join(fixtures_path, 'underscore.js'),
    )


def test_webasset_override_filters(tempdir, fixtures_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return fixtures_path

    @App.webasset_output()
    def get_output_path():
        return tempdir

    @App.webasset('jquery')
    def get_jquery_asset():
        yield 'jquery.js'

    @App.webasset_filter('js')
    def get_js_filter():
        return 'rjsmin'

    class DebugApp(App):
        pass

    @DebugApp.webasset_filter('js')
    def get_debug_js_filter():
        return None

    morepath.commit(DebugApp, App)

    bundles = list(App().config.webasset_registry.get_bundles('jquery'))
    assert len(bundles) == 1
    assert bundles[0].filters[0].name == 'rjsmin'

    bundles = list(DebugApp().config.webasset_registry.get_bundles('jquery'))
    assert len(bundles) == 1
    assert not bundles[0].filters


def test_webasset_override_filter_through_bundle(tempdir, fixtures_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return fixtures_path

    @App.webasset_output()
    def get_output_path():
        return tempdir

    @App.webasset('jquery')
    def get_jquery_asset():
        yield 'jquery.js'

    @App.webasset_filter('js')
    def get_js_filter():
        return 'rjsmin'

    class DebugApp(App):
        pass

    @DebugApp.webasset('common', filters={'js': None})
    def get_debug_js_filter():
        yield 'jquery'

    morepath.commit(DebugApp, App)

    bundles = list(App().config.webasset_registry.get_bundles('jquery'))
    assert len(bundles) == 1
    assert bundles[0].filters[0].name == 'rjsmin'

    bundles = list(DebugApp().config.webasset_registry.get_bundles('common'))
    assert len(bundles) == 1
    assert not bundles[0].filters


def test_webasset_mixed_bundles(tempdir, fixtures_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return fixtures_path

    @App.webasset_output()
    def get_output_path():
        return tempdir

    @App.webasset('common')
    def get_jquery_asset():
        yield 'jquery.js'
        yield 'extra.css'

    morepath.commit(App)

    bundles = list(App().config.webasset_registry.get_bundles('common'))
    assert len(bundles) == 2

    assert bundles[0].output.endswith('jquery.js.bundle.js')
    assert bundles[0].contents == (
        os.path.join(fixtures_path, 'jquery.js'),
    )

    assert bundles[1].output.endswith('extra.css.bundle.css')
    assert bundles[1].contents == (
        os.path.join(fixtures_path, 'extra.css'),
    )


def test_webasset_compiled_bundle(tempdir, fixtures_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return fixtures_path

    @App.webasset_output()
    def get_output_path():
        return tempdir

    @App.webasset_filter('scss')
    def get_scss_filter():
        return 'pyscss'

    @App.webasset('theme')
    def get_jquery_asset():
        yield 'main.scss'
        yield 'extra.css'

    morepath.commit(App)

    bundles = list(App().config.webasset_registry.get_bundles('theme'))
    assert len(bundles) == 2

    assert bundles[0].output.endswith('main.scss.bundle.css')
    assert bundles[0].contents == (
        os.path.join(fixtures_path, 'main.scss'),
    )

    assert bundles[1].output.endswith('extra.css.bundle.css')
    assert bundles[1].contents == (
        os.path.join(fixtures_path, 'extra.css'),
    )


def test_webasset_environment(tempdir, fixtures_path):

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_path():
        return fixtures_path

    @App.webasset_output()
    def get_output_path():
        return tempdir

    @App.webasset_filter('scss')
    def get_scss_filter():
        return 'pyscss'

    @App.webasset('js')
    def get_js_asset():
        yield 'jquery.js'
        yield 'underscore.js'

    @App.webasset('css')
    def get_css_asset():
        yield 'main.scss'
        yield 'extra.css'

    @App.webasset('common')
    def get_common_assets():
        yield 'js'
        yield 'css'

    morepath.commit(App)

    e = App().config.webasset_registry.get_environment()

    for asset in ('jquery.js', 'underscore.js'):
        assert e[asset].output.endswith('{}.bundle.js'.format(asset))
        assert e[asset].contents == (os.path.join(fixtures_path, asset),)
        assert len(e[asset].urls()) == 1

    for asset in ('main.scss', 'extra.css'):
        assert e[asset].output.endswith('{}.bundle.css'.format(asset))
        assert e[asset].contents == (os.path.join(fixtures_path, asset),)
        assert len(e[asset].urls()) == 1

    assert e['js'].output.endswith('js.bundle.js')
    assert e['js'].contents == (
        os.path.join(fixtures_path, 'jquery.js'),
        os.path.join(fixtures_path, 'underscore.js'),
    )

    assert len(e['js'].urls()) == 1

    assert e['css'].output.endswith('css.bundle.css')
    assert len(e['css'].contents) == 2
    assert e['css'].contents[0].output.endswith('main.scss.bundle.css')
    assert e['css'].contents[0].contents == (
        os.path.join(fixtures_path, 'main.scss'),
    )

    assert e['css'].contents[1].output.endswith('extra.css.bundle.css')
    assert e['css'].contents[1].contents == (
        os.path.join(fixtures_path, 'extra.css'),
    )

    assert len(e['css'].urls()) == 1

    assert e['common'].output.endswith('js.bundle.js')
    assert e['common'].contents == e['js'].contents
    assert len(e['common'].urls()) == 1

    assert e['common_1'].contents[0].output == e['css'].contents[0].output
    assert e['common_1'].contents[0].contents == e['css'].contents[0].contents
    assert e['common_1'].contents[1].output == e['css'].contents[1].output
    assert e['common_1'].contents[1].contents == e['css'].contents[1].contents
    assert len(e['common_1'].urls()) == 1



import atexit
import inspect
import os.path
import shutil
import tempfile

from dectate import Action
from webassets import Bundle, Environment


class Asset(object):
    """ Represents a registered asset which points to one or more files or
    child-assets.

    """

    __slots__ = ('name', 'assets', 'filters')

    def __init__(self, name, assets, filters):
        self.name = name
        self.assets = assets
        self.filters = filters

    def __eq__(self, other):
        return self.name == self.name \
            and self.assets == self.assets \
            and self.filters == self.filters

    @property
    def is_pure(self):
        """ Returns True if this asset is "pure".

        Pure assets are assets which consist of a single file or a set of
        files which share one common extension.

        """

        if self.is_single_file:
            return True

        extensions = {a.split('.')[-1] for a in self.assets}
        extensions |= {None for a in self.assets if '.' not in a}

        return len(extensions) == 1 and None not in extensions

    @property
    def is_single_file(self):
        """ Returns True if this repesents a single file asset. """
        return len(self.assets) == 1 and '.' in self.assets[0]

    @property
    def path(self):
        """ Returns the path to the single file asset if possible. """
        assert self.is_single_file
        return self.assets[0]

    @property
    def extension(self):
        """ Returns the extension of this asset if it's a pure asset. """
        if self.is_pure:
            return self.assets[0].split('.')[-1]


class WebassetRegistry(object):
    """ A registry managing webasset bundles registered through directives. """

    def __init__(self):

        #: A list of all paths which should be searched for files (in order)
        self.paths = []

        #: The default filters for extensions. Each extension has a webassets
        #: filter string associated with it. (e.g. {'js': 'rjsmin'})
        self.filters = {}

        #: :class:`Asset` objects keyed by their name
        self.assets = {}

        #: The output path for all bundles (a temporary directory by default)
        self.output_path = temporary_directory = tempfile.mkdtemp()
        atexit.register(shutil.rmtree, temporary_directory)

        #: A cache of created bundles
        self.cached_bundles = {}

        #: The url passed to the webasset environment
        self.url = 'assets'

        #: more.webasset only publishes js/css files - other file extensions
        #: need to be compiled into either and mapped accordingly
        self.mapping = {
            'coffee': 'js',
            'dust': 'js',
            'jst': 'js',
            'jsx': 'js',
            'less': 'css',
            'sass': 'css',
            'scss': 'css',
            'ts': 'js',
        }

    def register_path(self, path):
        """ Registers the given path as a path to be searched for files.

        The paths are prepended, so each new path has higher precedence than
        all the already registered paths.

        """
        assert os.path.isabs(path), "absolute paths only"
        self.paths.insert(0, os.path.normpath(path))

    def register_filter(self, name, filter):
        """ Registers a filter, overriding any existing filter of the same
        name.

        """
        self.filters[name] = filter

    def register_asset(self, name, assets, filters=None):
        """ Registers a new asset.

        """

        assert '.' not in name, "asset names may not contain dots ({})".format(
            name
        )

        # keep track of asset bundles
        self.assets[name] = Asset(
            name=name,
            assets=assets,
            filters=filters or self.filters
        )

        # and have one additional asset for each file
        for asset in assets:
            basename = os.path.basename(asset)

            # files are entries with an extension
            if '.' in basename:
                path = os.path.normpath(self.find_file(asset))

                self.assets[basename] = Asset(
                    name=basename,
                    assets=(path, ),
                    filters=filters or self.filters
                )
            else:
                assert asset in self.assets, "unknown asset {}".format(asset)

    def find_file(self, name):
        """ Searches for the given file by name using the current paths. """

        if os.path.isabs(name):
            return name

        searched = set()

        for path in self.paths:
            if path in searched:
                continue

            target = os.path.join(path, name)

            if os.path.isfile(target):
                return target

            searched.add(path)

        raise LookupError("Could not find {} in paths".format(name))

    def merge_filters(self, *filters):
        """ Takes a list of filters and merges them.

        The last filter has the highest precedence.

        """
        result = {}

        for filter in filters:
            if filter:
                result.update(filter)

        return result

    def get_bundles(self, name, filters=None):
        """ Yields all the bundles for the given name (an asset). """

        assert name in self.assets, "unknown asset {}".format(name)
        assert self.output_path, "no webasset_output path set"

        asset = self.assets[name]
        filters = self.merge_filters(self.filters, asset.filters, filters)

        if asset.is_pure:

            if asset.is_single_file:
                files = (asset.path, )
            else:
                files = (
                    a.path for a in (self.assets[a] for a in asset.assets))

            extension = self.mapping.get(asset.extension, asset.extension)
            assert extension in ('js', 'css')

            yield Bundle(
                *files,
                filters=filters.get(asset.extension),
                output='{}.bundle.{}'.format(name, extension)
            )
        else:
            for sub in (self.assets[a] for a in asset.assets):
                for bundle in self.get_bundles(sub.name, filters=filters):
                    yield bundle

    def get_environment(self):
        """ Returns the webassets environment, registering all the bundles. """

        debug = os.environ.get('MORE_WEBASSETS_DEBUG', '').lower().strip() in (
            'true', '1'
        )

        env = Environment(
            directory=self.output_path,
            load_path=self.paths,
            url=self.url,
            debug=debug
        )

        for asset in self.assets:
            bundles = tuple(self.get_bundles(asset))

            js = tuple(b for b in bundles if b.output.endswith('.js'))
            css = tuple(b for b in bundles if b.output.endswith('.css'))

            if js:
                js_bundle = len(js) == 1 and js[0] or Bundle(
                    *js, output='{}.bundle.js'.format(asset)
                )
            else:
                js_bundle = None

            if css:
                css_bundle = len(css) == 1 and css[0] or Bundle(
                    *css, output='{}.bundle.css'.format(asset)
                )
            else:
                css_bundle = None

            if js_bundle and css_bundle:
                js_bundle.next_bundle = asset + '_1'
                env.register(asset, js_bundle)
                env.register(asset + '_1', css_bundle)
            elif js_bundle:
                env.register(asset, js_bundle)
            else:
                env.register(asset, css_bundle)

        return env


class PathMixin(object):

    def absolute_path(self, path):
        if os.path.isabs(path):
            return path
        else:
            return os.path.join(os.path.dirname(self.code_info.path), path)


class WebassetPath(Action, PathMixin):
    """ Registers a path with more.webassets.

    Registered paths are searched for assets registered::

        @App.webasset_path()
        def get_asset_path():
            return 'assets/js'  # relative to the directory of the code file

        @App.webasset('jquery.js')
        def get_jquery_asset():
            yield 'jquery.js'  # expected to be at assets/js/jquery.js

    Registered paths can be accumulated, that is you can't override existing
    paths, you can just add new paths which take precedence
    (think ``PATH=/new/path:$PATH``).

    Therefore paths registered first are searched last and paths registered
    by a parent class are search after paths registered by the child class.

    """

    config = {
        'webasset_registry': WebassetRegistry
    }

    def identifier(self, webasset_registry):
        return object()

    def absolute_path(self, path):
        if os.path.isabs(path):
            return path
        else:
            return os.path.abspath(
                os.path.join(os.path.dirname(self.code_info.path), path)
            )

    def perform(self, obj, webasset_registry):
        path = self.absolute_path(obj())
        assert os.path.isdir(path), "'{}' does not exist".format(path)

        webasset_registry.register_path(self.absolute_path(obj()))


class WebassetOutput(Action, PathMixin):
    """ Sets the output path for all bundles.

    For example::

        @App.webasset_output()
        def get_output_path():
            return 'assets/bundles'

    """

    group_class = WebassetPath

    def identifier(self, webasset_registry):
        return self.__class__

    def perform(self, obj, webasset_registry):
        webasset_registry.output_path = self.absolute_path(obj())


class WebassetFilter(Action):
    """ Registers a default filter for an extension.

    Filters are strings interpreted by `webasset`::

        @App.webasset_filter('js')
        def get_js_filter():
            return 'rjsmin'

        @App.webasset_filter('scss')
        def get_scss_filter():
            return 'pyscss'

    For a list of available filters see
    `<http://webassets.readthedocs.org/en/latest/builtin_filters.html>`_.

    """

    group_class = WebassetPath

    def __init__(self, name):
        self.name = name

    def identifier(self, webasset_registry):
        return self.name

    def perform(self, obj, webasset_registry):
        webasset_registry.register_filter(self.name, obj())


class WebassetMapping(Action):
    """ Maps an extension to either css or js.

    You usually don't have to use this, as more.webassets comes with default
    values. If you do, please open an issue so your mapping may be added
    to more.webassets.

    Example::

        @App.webasset_mapping('jsx')
        def get_jsx_mapping():
            return 'js'

        @App.webasset_mapping('less')
        def get_jsx_mapping():
            return 'css'

    """

    group_class = WebassetPath

    def __init__(self, name):
        self.name = name

    def identifier(self, webasset_registry):
        return self.name

    def perform(self, obj, webasset_registry):
        webasset_registry.mapping[self.name] = obj()


class WebassetUrl(Action):
    """ Defines the url under which the bundles should be served.

    Passed to the webasset environment, this is basically a url path prefix::

        @App.webasset_url()
        def get_webasset_url():
            return 'my-assets'

    Defaults to 'assets'.

    """

    group_class = WebassetPath

    def identifier(self, webasset_registry):
        return self.__class__

    def perform(self, obj, webasset_registry):
        webasset_registry.url = obj()


class Webasset(Action):
    """ Registers an asset which may then be included in the page.

    For example::

        @App.webasset('tooltip')
        def get_tooltip_asset():
            yield 'tooltip.js'
            yield 'tooltip.css'

    Assets may be included by using
    :meth:`more.webassets.core.IncludeRequest.include`::

        @App.view(model=Model)
        def view_model(self, request):
            request.include('tooltip')

    Asset functions must be generators. They may include a mixed set of
    assets. So you can freely mix css, js, less and so on. When you include
    the asset by name those files are automatically put into appropriate
    webasset bundles.

    You may also define a custom filter which only applies to the registered
    assets (they override the default filters registerd by `webasset_filter`)::

        @App.webasset('tooltip', filters={'css': 'cssmin'})
        def get_tooltip_asset():
            yield 'tooltip.js'
            yield 'tooltip.css'

    Assets defined in a parent class may be overridden by using the same
    name, or they may be reused. This means you can have applications which
    define assets for you that you can reuse::

        @BaseApp.webasset('react')
        def get_react_asset():
            yield 'react.js'

        @App.webasset('widget')
        def get_widget_asset():
            yield 'react'
            yield 'widget.jsx'

    Note that webassets may not contain path separators. You're supposed to
    register all paths which should be searched, and then you only work
    with filenames.

    """

    depends = [
        WebassetFilter,
        WebassetMapping,
        WebassetOutput,
        WebassetPath,
        WebassetUrl
    ]
    group_class = WebassetPath

    def __init__(self, name, filters=None):
        self.name = name
        self.filters = filters

    def identifier(self, webasset_registry):
        return self.name

    def perform(self, obj, webasset_registry):
        assert inspect.isgeneratorfunction(obj), "webasset expects a generator"
        webasset_registry.register_asset(
            self.name, tuple(asset for asset in obj()), self.filters
        )



from more.webassets.tweens import InjectorTween, PublisherTween
from morepath.request import Request
from morepath.app import App
from dectate import directive
from ordered_set import OrderedSet
from . import directives


class IncludeRequest(Request):
    """ Adds the ability to include webassets bundles on the request.

    If the bundle does not exist, a KeyError will be raised during the
    rendering of the response, after the view has returned.

    Including a bundle multiple times will have the same result as
    including it once.

    The bundles are rendered in the order in which they were included. Bundles
    that are included first, are also rendered first.

    For example:

        @App.html(model=Model)
        def view(self, request):
            request.include('jquery')  # includes the jquery bundle

    """

    def __init__(self, *args, **kwargs):
        super(IncludeRequest, self).__init__(*args, **kwargs)
        self.included_assets = OrderedSet()

    def include(self, resource):
        self.included_assets.add(resource)


class WebassetsApp(App):
    """ Defines an app that servers webassets. """

    request_class = IncludeRequest

    webasset_path = directive(directives.WebassetPath)

    webasset_output = directive(directives.WebassetOutput)

    webasset_filter = directive(directives.WebassetFilter)

    webasset_mapping = directive(directives.WebassetMapping)

    webasset_url = directive(directives.WebassetUrl)

    webasset = directive(directives.Webasset)


@WebassetsApp.tween_factory()
def webassets_injector_tween(app, handler):
    """ Wraps the response with the injector and the publisher tween.

    See :class:`webassets.tweens.InjectorTween` and
    :class:`webassets.tweens.PublisherTween`.

    """

    env = app.config.webasset_registry.get_environment()

    injector_tween = InjectorTween(env, handler)
    publisher_tween = PublisherTween(env, injector_tween)

    return publisher_tween



import os
import time
import webob

from datetime import timedelta
from webob.static import FileApp

try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote


# content types and methods that get handled by the injector/publisher
CONTENT_TYPES = {'text/html', 'application/xhtml+xml'}
METHODS = {'GET', 'POST', 'HEAD'}

# arbitrarily define forever as 10 years in the future
FOREVER = timedelta(days=365 * 10).total_seconds()


# what separators does this operating system provide that are not a slash?
_os_alt_seps = set(sep for sep in [os.path.sep, os.path.altsep]
                   if sep not in (None, '/'))

# what elements may not be in a path element?
_insecure_elements = set(['..', '.', '']) | _os_alt_seps


def is_subpath(directory, path):
    """ Returns true if the given path is inside the given directory. """
    directory = os.path.join(os.path.realpath(directory), '')
    path = os.path.realpath(path)

    # return true, if the common prefix of both is equal to directory
    # e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
    return os.path.commonprefix([path, directory]) == directory


def has_insecure_path_element(path):
    """ Returns true if the given path contains an insecure path element.
    That is '..' or '.' or ''
    """
    if _insecure_elements.intersection(path.split('/')):
        return True

    return False


class InjectorTween(object):
    """ Injects the webasset urls into the response. """

    def __init__(self, environment, handler):
        self.environment = environment
        self.handler = handler
        self._urls = {}

    def urls_by_resource(self, resource):
        if self.environment.debug or resource not in self._urls:
            self._urls[resource] = []

            bundle = self.environment[resource]
            while bundle is not None:
                self._urls[resource].extend(bundle.urls())

                try:
                    bundle = self.environment[getattr(bundle, 'next_bundle')]
                except (AttributeError, KeyError):
                    bundle = None

        return self._urls[resource]

    def urls_to_inject(self, request, suffix=None):
        for resource in request.included_assets:
            for url in self.urls_by_resource(resource):
                filename = url.split('?')[0]

                if suffix and not filename.endswith(suffix):
                    continue

                yield '/' + url

    def __call__(self, request):
        response = self.handler(request)

        if request.method not in METHODS:
            return response

        if not response.content_type:  # may be None if the code is != 200
            return response

        if response.content_type.lower() not in CONTENT_TYPES:
            return response

        scripts = '\n'.join(
            '<script type="text/javascript" src="{}"></script>'.format(url)
            for url in self.urls_to_inject(request, '.js')
        )

        stylesheets = '\n'.join(
            '<link rel="stylesheet" type="text/css" href="{}">'.format(url)
            for url in self.urls_to_inject(request, '.css')
        )

        if scripts:
            response.body = response.body.replace(
                b'</body>', scripts.encode('utf-8') + b'</body>', 1
            )

        if stylesheets:
            response.body = response.body.replace(
                b'</head>', stylesheets.encode('utf-8') + b'</head>', 1
            )

        return response


class PublisherTween(object):
    """ Returns the webassets if the request begins with the
    :attr:`WebassetsApp.webassets_url`.

    """

    def __init__(self, environment, handler):
        self.environment = environment
        self.handler = handler

    def __call__(self, request):
        publisher_signature = request.path_info_peek()

        if publisher_signature != self.environment.url:
            return self.handler(request)

        subpath = request.path_info.replace(publisher_signature, '').strip('/')
        subpath = unquote(subpath)

        if has_insecure_path_element(subpath):
            return webob.exc.HTTPNotFound()

        asset = os.path.join(self.environment.directory, subpath)
        asset = os.path.abspath(asset)

        # I'm not entirely at ease with loading a file from disk and returning
        # it over the web. So as an extra precaution I want to make sure
        # that only files *inside* the assets folder will be served.
        #
        # I doubt webob let's this happen, but if somebody is ever able to
        # create an url resulting in a path that points outside the assets
        # directory this check might help.
        if not is_subpath(self.environment.directory, asset):
            return webob.exc.HTTPNotFound()

        # This is possibly too paranoid
        if os.path.islink(asset):
            return webob.exc.HTTPNotFound()

        if not os.path.isfile(asset):
            return webob.exc.HTTPNotFound()

        response = request.get_response(FileApp(asset))

        if response.status_code == 200:
            response.cache_control.max_age = FOREVER
            response.expires = time.time() + FOREVER

        return response



import more.webassets
import morepath
import os
import pytest

from datetime import datetime
from more.webassets import WebassetsApp
from more.webassets.tweens import is_subpath, has_insecure_path_element
from webtest import TestApp as Client


def prepare_fixtures(directory):
    os.mkdir(os.path.join(directory, 'common'))
    os.mkdir(os.path.join(directory, 'output'))
    os.mkdir(os.path.join(directory, 'theme'))

    # if those javascripts are changed, the tests below need to be updated
    # with the correct md5 hashes
    with open(os.path.join(directory, 'common', 'jquery.js'), 'w') as f:
        f.write("""
            /* fake jquery */
            var $ = function(){};
        """)

    with open(os.path.join(directory, 'common', 'underscore.js'), 'w') as f:
        f.write("""
            /* fake underscore */
            var _ = function(){};
        """)

    with open(os.path.join(directory, 'theme', 'main.scss'), 'w') as f:
        f.write("""
            body {
                a {
                    color: blue;
                }
            }
        """)

    with open(os.path.join(directory, 'theme', 'other.css'), 'w') as f:
        f.write("""
            h1 {
                font-size: 2rem;
            }
        """)

    with open(os.path.join(directory, 'common', 'extra.js'), 'w') as f:
        f.write("""
            $(document).ready(function(){});
        """)


def spawn_test_app(tempdir):
    prepare_fixtures(tempdir)

    html = """
        <html>
            <head></head>
            <body>
                <p>Bar</p>
            </body>
        </html>
    """

    def render_plain(content, request):
        response = morepath.Response(content)
        response.content_type = 'text/plain'
        return response

    class App(WebassetsApp):
        pass

    @App.webasset_path()
    def get_default_assets_path():
        return os.path.join(tempdir, 'common')

    @App.webasset_path()
    def get_theme_assets_path():
        return os.path.join(tempdir, 'theme')

    @App.webasset_output()
    def get_output_path():
        return os.path.join(tempdir, 'output')

    @App.webasset_filter('js')
    def get_js_filter():
        return 'rjsmin'

    @App.webasset_filter('scss')
    def get_scss_filter():
        return 'pyscss'

    @App.webasset(name='common')
    def get_common_assets():
        yield 'jquery.js'
        yield 'underscore.js'

    @App.webasset(name='extra')
    def get_extra_asset():
        yield 'extra.js'

    @App.webasset(name='theme')
    def get_theme_asset():
        yield 'main.scss'

    @App.path('')
    class Root(object):
        pass

    @App.html(model=Root)
    def index(self, request):
        bundle = request.params.get('bundle')

        if bundle:
            request.include(bundle)

        return html

    @App.view(model=Root, name='plain', render=render_plain)
    def plain(self, request):
        request.include('common')
        return html

    @App.html(model=Root, name='put', request_method='PUT')
    def put(self, request):
        request.include('common')
        return html

    @App.html(model=Root, name='alljs')
    def alljs(self, request):
        request.include('common')
        request.include('extra')
        return html

    morepath.scan(more.webassets)
    morepath.commit(App)

    return App()


def test_inject_webassets(tempdir):
    client = Client(spawn_test_app(tempdir))

    assert '<head></head>' in client.get('/').text

    with pytest.raises(KeyError):
        assert '<head></head>' in client.get('?bundle=inexistant').text

    # the version of these assets stays the same because it's basically the
    # md5 hash of the javascript files included
    injected_html = (
        '<script type="text/javascript" '
        'src="/assets/common.bundle.js?ddc71aa3"></script></body>')

    assert injected_html in client.get('?bundle=common').text

    injected_html = (
        '<link rel="stylesheet" type="text/css" '
        'href="/assets/theme.bundle.css?32fda411"></head>')

    assert injected_html in client.get('?bundle=theme').text

    page = client.get('/alljs').text
    assert page.find('common.bundle.js') < page.find('extra.bundle.js')


def test_publish_webassets(tempdir):
    client = Client(spawn_test_app(tempdir))

    url = '/assets/common.bundle.js?ddc71aa3'

    # before the urls() have not been called by the injector, the bundles
    # won't have been created
    assert client.get(url, expect_errors=True).status_code == 404

    client.get('?bundle=common')

    assert client.get(url).text == 'var $=function(){};var _=function(){};'
    assert client.get(url).expires.year == datetime.utcnow().year + 10
    assert client.get(url).content_type == 'text/javascript'

    # do the same for css
    url = '/assets/theme.bundle.css?32fda411'

    assert client.get(url, expect_errors=True).status_code == 404

    client.get('?bundle=theme')

    assert client.get(url).text == 'body a {\n  color: blue; }\n'
    assert client.get(url).expires.year == datetime.utcnow().year + 10
    assert client.get(url).content_type == 'text/css'


def test_webassets_unhandled_content_type(tempdir):
    client = Client(spawn_test_app(tempdir))

    assert '<head></head>' in client.get('/plain').text


def test_webassets_unhandled_request_method(tempdir):
    client = Client(spawn_test_app(tempdir))

    assert '<head></head>' in client.put('/put').text


def test_is_subpath(tempdir):
    assert is_subpath('/', '/test')
    assert is_subpath('/asdf', '/asdf/asdf')
    assert not is_subpath('/asdf/', '/asdf')
    assert not is_subpath('/a', '/b')
    assert not is_subpath('/a', '/a/../b')


def test_insecure_path_element():
    assert has_insecure_path_element('../test.txt')
    assert has_insecure_path_element('./test.txt')
    assert has_insecure_path_element('/test.txt')
    assert not has_insecure_path_element('test.txt')
    assert not has_insecure_path_element('asdf/asdf/test.txt')



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

name = 'more.webassets'
description = (
    'An opinionated Webassets integration for Morepath.'
)
version = '0.3.4'


def get_long_description():
    import io
    readme = io.open('README.rst', encoding='utf-8').read()
    history = io.open('HISTORY.rst', encoding='utf-8').read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, history))


setup(
    name=name,
    version=version,
    description=description,
    long_description=get_long_description(),
    url='http://github.com/morepath/more.webassets',
    author='Seantis GmbH',
    author_email='info@seantis.ch',
    license='BSD',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=name.split('.')[:-1],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'morepath>=0.16',
        'ordered-set',
        'webassets',
        'webob'
    ],
    extras_require=dict(
        test=[
            'coverage',
            'pytest',
            'webtest',
            'pyscss'
        ],
    ),
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: BSD License',
    ]
)
